""" This module implements the command line interface for telewater,
using the modern and robust `typer`.
"""

import logging
import os
from typing import Optional

import typer
from dotenv import load_dotenv

from telewater import __version__
from telewater.main import start_bot

load_dotenv(".env")

FAKE = bool(os.getenv("FAKE"))

app = typer.Typer(add_completion=False)


def version_callback(value: bool):
    if value:
        print(__version__)
        raise typer.Exit()


def verbosity_callback(value: bool):
    if value:
        level = logging.INFO
    else:
        level = logging.WARNING
    logging.basicConfig(level=level)
    logging.info("Verbosity turned on. \nThis is suitable for debugging.\n")


@app.command()
def main(
    API_ID: int = typer.Option(
        ...,
        "--API_ID",
        help="API ID obtained from my.telegram.org",
        envvar="API_ID",
        prompt="Paste your API ID (input hidden)",
        hide_input=True,
    ),
    API_HASH: str = typer.Option(
        ...,
        "--API_HASH",
        help="API HASH obtained from my.telegram.org",
        envvar="API_HASH",
        prompt="Paste your API HASH (input hidden)",
        hide_input=True,
    ),
    name: str = typer.Option(
        ...,
        "--name",
        "-n",
        help="Name of the bot you are runnning.",
        envvar="BOT_USERNAME",
        prompt="Enter the bot name",
    ),
    token: str = typer.Option(
        ...,
        "--token",
        "-t",
        help="Bot Token obtained from @BotFather.",
        envvar="BOT_TOKEN",
        prompt="Paste the bot token (input hidden)",
        hide_input=True,
        confirmation_prompt=True,
    ),
    verbose: Optional[bool] = typer.Option(
        None,
        "--loud",
        "-l",
        callback=verbosity_callback,
        envvar="LOUD",
        help="Increase output verbosity.",
    ),
    version: Optional[bool] = typer.Option(
        None,
        "--version",
        "-v",
        callback=version_callback,
        help="Show version and exit.",
    ),
):
    """A telegram bot that applies watermark on images, gifs and videos.

    For detailed docs read : https://github.com/aahnik/telewater
    """

    if FAKE:
        print(f"name is {name} and token is {token}")
        print(f"API_ID = {API_ID} and API_HASH = {API_HASH}")
        quit(1)

        # when the env var FAKE is truthy, then no real work is done
        # this is for CLI testing purposes

    start_bot(API_ID, API_HASH, name, token)


# AAHNIK 2021



""" Various utility functions are defined in this module.
"""


import shutil

import requests

# TODO: make file download using asynchronous method
# import aiohttp
# import aiofiles

# async def download_image(url: str, filename: str) -> bool:
#     async with aiohttp.ClientSession() as session:
#         async with session.get(url) as resp:
#             if resp.status == 200:
#                 f = await aiofiles.open(filename, mode='wb')
#                 await f.write(await resp.read())
#                 await f.close()


def download_image(url: str, filename: str = "image.png") -> bool:
    try:
        print("Downloading image ... ")
        response = requests.get(url, stream=True)
        if response.status_code == 200:
            print("Got file response")
            with open(filename, "wb") as file:
                response.raw.decode_content = True
                shutil.copyfileobj(response.raw, file)
    except Exception as err:
        print(err)
        return False
    else:
        print("File created image")
        return True


def get_args(text: str):
    splitted = text.split(" ", 1)
    if not len(splitted) == 2:
        raise ValueError(f"No args provided to command.")
    else:
        prefix, args = splitted
    print(prefix)
    args = args.strip()
    print(args)
    return args



""" This module defines the constants or default values.
"""
from pydantic import BaseModel, validator


class Config(BaseModel):
    watermark: str = "https://user-images.githubusercontent.com/66209958/109513526-35883200-7acb-11eb-97ed-c0b2ca72119a.png"
    x_off: int = 10
    y_off: int = 10
    frame_rate: int = 15
    preset: str = "ultrafast"

    @validator("preset")
    def validate_preset(val):
        allowed = ["ultrafast", "fast", "medium", "slow"]
        if not val in allowed:
            raise ValueError(f"Choose preset from {allowed}")
        return val


HELP = """
This bot is made with free and open source code.

Please star on GitHub

https://github.com/aahnik/telewater

"""

COMMANDS = {
    "start": "start the bot or check if alive",
    "setimg": "set the image for applying as watermark",
    "setpos": "set the position to apply the watermark",
    "help": "learn how to use the bot",
}

config = Config()



""" This module provides the pythonic entry point for accessing telewater.
"""


import os

from telethon import TelegramClient

from telewater.bot import ALL_EVENTS
from telewater.const import config
from telewater.utils import download_image


def start_bot(API_ID: int, API_HASH: str, name: str, token: str):
    os.makedirs(name, exist_ok=True)
    os.chdir(name)

    if config.watermark:
        download_image(url=config.watermark)

    client = TelegramClient(name, API_ID, API_HASH).start(bot_token=token)

    for key, val in ALL_EVENTS.items():
        print(f"Adding event {key}")
        client.add_event_handler(*val)

    print(f"Started bot {name}")
    client.run_until_disconnected()



""" Watermark is applied here."""

import os

from telewater.const import config


def watermark_video(video_file: str):
    """Apply watermark to video or gifs"""
    output_file = f"watered_{video_file}"
    # TODO: allow time stamping in file names
    # TODO: allow the ffmpeg options to be customized by bot admins
    # TODO: allow standard watermark positions like TOP/MIDDLE/BOTTOM/LEFT/RIGHT in total 9 combinations
    command = f'ffmpeg -i {video_file} \
        -i image.png \
        -an -dn -sn -r {config.frame_rate} \
        -preset {config.preset} \
        -tune zerolatency  -tune fastdecode \
        -filter_complex "overlay={config.x_off}:{config.y_off}" \
        {output_file}'

    print(command)
    os.system(command)
    return output_file


def watermark_image(image_file: str):
    pass



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['telewater']

package_data = \
{'': ['*']}

install_requires = \
['Pillow>=8.1.2,<9.0.0',
 'PyYAML>=5.4.1,<6.0.0',
 'Telethon>=1.21.1,<2.0.0',
 'aiohttp>=3.7.4,<4.0.0',
 'cryptg>=0.2.post2,<0.3',
 'hachoir>=3.1.2,<4.0.0',
 'pydantic>=1.8.1,<2.0.0',
 'python-dotenv>=0.16.0,<0.17.0',
 'requests>=2.25.1,<3.0.0',
 'typer>=0.3.2,<0.4.0']

entry_points = \
{'console_scripts': ['telewater = telewater.cli:app']}

setup_kwargs = {
    'name': 'telewater',
    'version': '0.0.10',
    'description': 'A telegram bot that applies watermark on images, gifs and videos.',
    'long_description': '# telewater\n\nA telegram bot that applies watermark on images, gifs and videos.\n\n## Features\n\n- **Fast** because it is made using async libraries.\n- **Simple** to use.\n- Any one who uses an instance of the bot will have to use the same watermark and position. This is meant to be used by **single person/organization** (by only you or your team), as configuration is global.\n- **No database** connection required.\n- It **does not store media** (photos/videos/gifs) on the server. Media is deleted immediately after the watermarked version is sent to the user.\n\n\n## Installation\n\nIf you are familiar with **Docker** then [click here](https://github.com/aahnik/telewater/wiki/Install-and-run-using-docker) otherwise, continue reading.\n\n### Requirements\n\nMake sure to have these installed in your system.\n\n- [python3.9+](https://www.python.org/) and [pip](https://pip.pypa.io/en/stable/installing/) (the bot is built with the telethon library)\n- [ffmpeg](https://ffmpeg.org/) (used by the bot for applying watermark)\n\n### Verification\n\nOpen you terminal to check if you have all basic requirements properly installed.\n\n1. Run `python --version` and you should get something like this `Python 3.9.2` (or above).\n2. Run `pip --version` and you should get `pip 20.2.2` (or above).\n\n    > Some systems may require to use `python3` and `pip3` instead of the above.\n\n3. Run `ffmpeg -h` and it should display a help message and version above `4.2.4`.\n\n### Install via pip\n\n```shell\n$ pip install telewater\n```\n\n\n## Usage\n\nTelewater has a simple command line interface to start the bot.\n\nSimply open your terminal and run `telewater`. It will prompt you to enter the required information.\n\n\n## Further reading\n\n- [Environment Variables](https://github.com/aahnik/telewater/wiki/Environment-Variables)\n- [Telewater CLI usage](https://github.com/aahnik/telewater/wiki/Telewater-CLI-usage)\n- [Install and run using docker](https://github.com/aahnik/telewater/wiki/Install-and-run-using-docker)\n- [Deploy to Digital Ocean](https://github.com/aahnik/telewater/wiki/Deploy-to-Digital-Ocean)\n- [Run multiple instances](https://github.com/aahnik/telewater/wiki/Run-multiple-instances)\n\n\nFor any further help, feel free to [create an issue](https://github.com/aahnik/telewater/issues) in the GitHub repo.\n\n\n<!-- AAHNIK 2021 -->\n',
    'author': 'aahnik',
    'author_email': 'daw@aahnik.dev',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/aahnik/telewater',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.9,<4.0',
}


setup(**setup_kwargs)



""" A telegram bot that applies watermark on images, gifs and videos. """

from importlib.metadata import version

from telewater.main import start_bot

__version__ = version(__package__)

# AAHNIK 2021



""" This module defines the functions that handle different events.
"""

import logging
import os

from telethon import TelegramClient, events

from telewater.const import HELP, Config, config
from telewater.utils import download_image, get_args
from telewater.watermark import watermark_video


async def start(event):
    await event.respond("Hi! I am alive.")
    raise events.StopPropagation


async def bot_help(event):
    try:
        await event.respond(HELP)
    finally:
        raise events.StopPropagation


async def set_config(event):
    """usage /set KEY: VAL"""
    global config
    try:
        pos_arg = get_args(event.message.text)
        splitted = pos_arg.split(":", 1)

        if not len(splitted) == 2:
            raise ValueError("Incorrect argument format")

        key, value = [item.strip() for item in splitted]

        config_dict = config.dict()
        if not key in config_dict.keys():
            raise ValueError(f"The key {key} is not a valid key in configuration.")

        config_dict[key] = value
        print(config_dict)

        config = Config(**config_dict)

        print(config)
        if key == "watermark":
            download_image(url=value)
        await event.respond(f"The value of {key} was set to {value}")

    except ValueError as err:
        print(err)
        await event.respond(str(err))
    except Exception as err:
        print(err)

    finally:
        raise events.StopPropagation


async def get_config(event):
    """usage /get KEY"""
    global config
    try:
        key = get_args(event.message.text)
        config_dict = config.dict()
        await event.respond(f"{config_dict.get(key)}")
    except ValueError as err:
        print(err)
        await event.respond(str(err))

    finally:

        raise events.StopPropagation


async def watermarker(event):
    # TODO: reject large files (above certain file limit)
    # TODO: also watermark photos
    global config
    if event.gif or event.video:

        mp4_file = await event.download_media("")
        # TODO: suffix the downloaded media with time-stamp and user id

        outf = watermark_video(mp4_file)
        print(outf)
        await event.client.send_file(event.sender_id, outf)
        os.remove(mp4_file)
        os.remove(outf)
    elif event.photo:
        await event.respond("Photos are currently not supported")


# TODO: fetch information about bot name
# TODO:set the bot commands

# client(functions.bots.SetBotCommandsRequest(
#     commands=[types.BotCommand(
#         command='some string here',
#         description='some string here'
#     )]
# ))
# client.run_until_disconnected()


ALL_EVENTS = {
    "start": (start, events.NewMessage(pattern="/start")),
    "help": (bot_help, events.NewMessage(pattern="/help")),
    "set": (set_config, events.NewMessage(pattern="/set")),
    "get": (get_config, events.NewMessage(pattern="/get")),
    "watermarker": (watermarker, events.NewMessage()),
}
# this is a dictionary where the keys are the unique string identifier for the events
# the values are a tuple consisting of callback function and event
